#!/usr/bin/python
# vpn.py: a top level interface to commands to run VPNs on a VM as
# different clients. The use case is allowing us to measure from more
# places.

import argparse
import logging
from random import shuffle
import os

import centinel.backend
import centinel.client
import centinel.config
import centinel.vpn.openvpn as openvpn
import centinel.vpn.hma as hma


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--auth-file', '-u', dest='auth_file', default=None,
                        help=("File with HMA username on first line, \n"
                              "HMA password on second line"))
    parser.add_argument('--crt-file', '-r', dest='crt_file', default=None,
                        help=("Certificate file for the current vpn\n"
                              "provider, if provided"))
    parser.add_argument('--create-hma-configs', dest='create_HMA',
                        action="store_true",
                        help='Create the openvpn config files for HMA')
    parser.add_argument('--shuffle', '-s', dest='shuffle_lists',
                        action="store_true",
                        help='Randomize the order of vantage points')
    parser.add_argument('--exclude', '-e', dest='exclude_list', default=None,
                        help=('Countries to exclude when scanning (comma '
                              'separated two letter country codes)'))
    parser.add_argument('--log-file', '-l', dest='log_file', default=None,
                        help="Log file location")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--directory', '-d', dest='directory',
                       help='Directory with experiments, config files, etc.')
    create_conf_help = ('Create configuration files for the given '
                        'openvpn config files so that we can treat each '
                        'one as a client. The argument should be a '
                        'directory with a subdirectory called openvpn '
                        'that contains the openvpn config files')
    group.add_argument('--create-config', '-c', help=create_conf_help,
                       dest='create_conf_dir')
    return parser.parse_args()


def scan_vpns(directory, auth_file, crt_file, exclude_list, shuffle_lists=False):
    """For each VPN, check if there are experiments and scan with it if
    necessary

    Note: the expected directory structure is
    args.directory
    -----vpns (contains the OpenVPN config files
    -----configs (contains the Centinel config files)
    -----exps (contains the experiments directories)

    """

    logging.info("Starting to run the experiments for each VPN")
    logging.warn("Excluding vantage points from: %s" % (exclude_list))

    # iterate over each VPN
    vpn_dir = return_abs_path(directory, "vpns")
    conf_dir = return_abs_path(directory, "configs")
    if auth_file is not None:
        auth_file = return_abs_path(directory, auth_file)
    if crt_file is not None:
        crt_file = return_abs_path(directory, crt_file)
    conf_list = os.listdir(conf_dir)

    if shuffle_lists:
        shuffle(conf_list)

    number = 1
    total = len(conf_list)

    for filename in conf_list:
        logging.info("Moving onto (%d/%d) %s" % (number, total, filename))
        print "(%d/%d) %s" % (number, total, filename)

        number += 1
        vpn_config = os.path.join(vpn_dir, filename)
        centinel_config = os.path.join(conf_dir, filename)

        # before starting the VPN, check if there are any experiments
        # to run
        config = centinel.config.Configuration()
        config.parse_config(centinel_config)

        # assuming that each VPN config file has a name like:
        # [ip-address].ovpn, we can extract IP address from filename
        # and use it to geolocate and fetch experiments before connecting
        # to VPN.
        vpn_address, extension = os.path.splitext(filename)
        country = None
        try:
            meta = centinel.backend.get_meta(config.params,
                                             vpn_address)
            if 'country' in meta:
                country = meta['country']
        except Exception as exp:
            logging.exception("%s: Failed to geolocate "
                              "%s: %s" % (filename, vpn_address, exp))

        if country and exclude_list and country in exclude_list:
            logging.info("%s: Skipping this server (%s)" % (filename, country))
            continue

        # try setting the VPN info (IP and country) to get appropriate
        # experiemnts and input data.
        try:
            centinel.backend.set_vpn_info(config.params, vpn_address, country)
        except Exception as exp:
            logging.exception("%s: Failed to set VPN info: %s" % (filename, exp))

        logging.info("%s: Synchronizing." % (filename))
        try:
            centinel.backend.sync(config.params)
        except Exception as exp:
            logging.exception("%s: Failed to sync: %s" % (filename, exp))

        if not experiments_available(config.params):
            logging.info("%s: No experiments available." % (filename))
            try:
                centinel.backend.set_vpn_info(config.params, vpn_address, country)
            except Exception as exp:
                logging.exception("Failed to set VPN info: %s" % (exp))
            continue

        logging.info("%s: Starting VPN." % (filename))

        vpn = openvpn.OpenVPN(timeout=60, auth_file=auth_file,
                              config_file=vpn_config, crt_file=crt_file)

        vpn.start()
        if not vpn.started:
            vpn.stop()
            logging.error("%s: Failed to start VPN!" % (filename))
            continue

        logging.info("%s: Running Centinel." % (filename))
        try:
            client = centinel.client.Client(config.params)
            centinel.conf = config.params
            client.setup_logging()
            client.run()
        except Exception as exp:
            logging.exception("%s: Error running Centinel: %s" % (filename, exp))

        logging.info("%s: Stopping VPN." % (filename))
        vpn.stop()

        logging.info("%s: Synchronizing." % (filename))
        try:
            centinel.backend.sync(config.params)
        except Exception as exp:
            logging.exception("%s: Failed to sync: %s" % (filename, exp))

        # try setting the VPN info (IP and country) to the correct address
        # after sync is over.
        try:
            centinel.backend.set_vpn_info(config.params, vpn_address, country)
        except Exception as exp:
            logging.exception("Failed to set VPN info: %s" % (exp))

def return_abs_path(directory, path):
    """Unfortunately, Python is not smart enough to return an absolute
    path with tilde expansion, so I writing functionality to do this

    """
    if directory is None or path is None:
        return
    directory = os.path.expanduser(directory)
    return os.path.abspath(os.path.join(directory, path))

def create_config_files(directory):

    """For each VPN file in directory/vpns, create a new configuration
    file and all the associated directories

    Note: the expected directory structure is
    args.directory
    -----vpns (contains the OpenVPN config files
    -----configs (contains the Centinel config files)
    -----exps (contains the experiments directories)
    -----results (contains the results)

    """
    logging.info("Starting to create config files from openvpn files")

    vpn_dir = return_abs_path(directory, "vpns")
    conf_dir = return_abs_path(directory, "configs")
    os.mkdir(conf_dir)
    home_dirs = return_abs_path(directory, "home")
    os.mkdir(home_dirs)
    for filename in os.listdir(vpn_dir):
        configuration = centinel.config.Configuration()
        # setup the directories
        home_dir = os.path.join(home_dirs, filename)
        os.mkdir(home_dir)
        configuration.params['user']['centinel_home'] = home_dir
        exp_dir = os.path.join(home_dir, "experiments")
        os.mkdir(exp_dir)
        configuration.params['dirs']['experiments_dir'] = exp_dir
        data_dir = os.path.join(home_dir, "data")
        os.mkdir(data_dir)
        configuration.params['dirs']['data_dir'] = data_dir
        res_dir = os.path.join(home_dir, "results")
        os.mkdir(res_dir)
        configuration.params['dirs']['results_dir'] = res_dir

        log_file = os.path.join(home_dir, "centinel.log")
        configuration.params['log']['log_file'] = log_file
        login_file = os.path.join(home_dir, "login")
        configuration.params['server']['login_file'] = login_file
        configuration.params['user']['is_vpn'] = True

        configuration.params['server']['verify'] = False
        configuration.params['experiments']['tcpdump_params'] = ["-i", "tun0"]
        # set vpn server to be sbu server by default
        configuration.params['server']['server_url'] = "https://130.245.145.7:8082"

        conf_file = os.path.join(conf_dir, filename)
        configuration.write_out_config(conf_file)


def experiments_available(config):
    logging.info("Starting to check for experiments with %s",
                 config['server']['server_url'])
    try:
        client = centinel.client.Client(config)
        if client.has_experiments_to_run():
            return True
    except Exception, exp:
        logging.exception("Unable to check schedule: %s", str(exp))

    return False


def run():
    """Entry point for all uses of centinel"""

    args = parse_args()
    logging.basicConfig(filename=args.log_file,
                        format="%(levelname)s %(asctime)s: %(message)s",
                        level=logging.INFO)
    if args.create_conf_dir:
        if args.create_HMA:
            hmaDir = return_abs_path(args.create_conf_dir, "vpns")
            hma.create_config_files(hmaDir)
        # create the config files for the openvpn config files
        create_config_files(args.create_conf_dir)
    else:
        scan_vpns(args.directory, args.auth_file, args.crt_file,
                  args.exclude_list, args.shuffle_lists)

if __name__ == "__main__":
    run()



from base64 import b64encode
import dns.rdatatype
import dns.message
import dns.resolver
import logging
import select
import socket
import threading
import time


def get_ips(host, nameserver=None, record="A"):
    nameservers = []
    if nameserver is not None:
        nameservers = [nameserver]
    return lookup_domain(host, nameservers=nameservers, rtype=record)


def lookup_domain(domain, nameservers=[], rtype="A", timeout=2):
    """Wrapper for DNSQuery method"""
    dns_exp = DNSQuery(domains=[domain], nameservers=nameservers, rtype=rtype,
                       timeout=timeout)
    return dns_exp.lookup_domain(domain)


def lookup_domains(domains, nameservers=[], rtype="A", timeout=10):
    dns_exp = DNSQuery(domains=domains, nameservers=nameservers, rtype=rtype,
                       timeout=timeout)
    return dns_exp.lookup_domains()


def send_chaos_queries():
    dns_exp = DNSQuery()
    return dns_exp.send_chaos_queries()


class DNSQuery():
    """Class to store state for all of the DNS queries"""

    def __init__(self, domains=[], nameservers=[], rtype="A", timeout=10,
                 max_threads=100):
        """Constructor for the DNS query class

        Params:
        nameserver- the nameserver to use, defaults to the local resolver
        rtype- the record type to lookup (as text), by default A
        timeout- how long to wait for a response, by default 10 seconds

        """
        self.domains = domains
        self.rtype = rtype
        self.timeout = timeout
        self.max_threads = max_threads
        if nameservers == []:
            nameservers = dns.resolver.Resolver().nameservers
        self.nameservers = nameservers
        self.results = {}
        self.threads = []

    def send_chaos_queries(self):
        """Send chaos queries to identify the DNS server and its manufacturer

        Note: we send 2 queries for BIND stuff per RFC 4892 and 1
        query per RFC 6304

        Note: we are not waiting on a second response because we
        shouldn't be getting injected packets here

        """
        names = ["HOSTNAME.BIND", "VERSION.BIND", "ID.SERVER"]
        self.results = {'exp-name': "chaos-queries"}
        for name in names:
            self.results[name] = {}
            for nameserver in self.nameservers:
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.settimeout(self.timeout)
                query = dns.message.make_query(name,
                                               dns.rdatatype.from_text("TXT"),
                                               dns.rdataclass.from_text("CH"))
                sock.sendto(query.to_wire(), (nameserver, 53))
                reads, _, _ = select.select([sock], [], [], self.timeout)
                if reads == []:
                    self.results[name][nameserver] = None
                else:
                    response = reads[0].recvfrom(4096)[0]
                    self.results[name][nameserver] = b64encode(response)
        return self.results

    def lookup_domains(self):
        """More complex DNS primitive that lookups domains concurrently

        Note: if you want to lookup multiple domains, you should use
        this function
        """
        thread_error = False
        thread_wait_timeout = 200
        ind = 1
        total_item_count = len(self.domains)
        for domain in self.domains:
            for nameserver in self.nameservers:
                wait_time = 0
                while threading.active_count() > self.max_threads:
                    time.sleep(1)
                    wait_time += 1
                    if wait_time > thread_wait_timeout:
                        thread_error = True
                        break

                if thread_error:
                    self.results["error"] = "Threads took too long to finish."
                    break
                log_prefix = "%d/%d: " % (ind, total_item_count)
                thread = threading.Thread(target=self.lookup_domain,
                                          args=(domain, nameserver,
                                                log_prefix))
                thread.setDaemon(1)
                thread.start()
                self.threads.append(thread)
            if thread_error:
                break
            ind += 1

        for thread in self.threads:
            thread.join(self.timeout * 3)
        return self.results

    def lookup_domain(self, domain, nameserver=None, log_prefix = ''):
        """Most basic DNS primitive that lookups a domain, waits for a
        second response, then returns all of the results

        Params:
        domain- the domain to lookup
        nameserver- the nameserver to use

        Note: if you want to lookup multiple domains you *should not* use
        this function, you should use lookup_domains because this does
        blocking IO to wait for the second response

        """
        # get the resolver to use
        if nameserver is None:
            nameserver = self.nameservers[0]
        results = {'domain': domain, 'nameserver': nameserver}
        # construct the socket to use
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.settimeout(self.timeout)

        logging.debug("%sQuerying DNS enteries for "
                      "%s (nameserver: %s)." % (log_prefix, domain, nameserver))

        # construct and send the request
        request = dns.message.make_query(domain,
                                         dns.rdatatype.from_text(self.rtype))
        results['request'] = b64encode(request.to_wire())
        sock.sendto(request.to_wire(), (nameserver, 53))

        # read the first response from the socket
        reads, _, _ = select.select([sock], [], [], self.timeout)
        # if we didn't get anything, then set the results to nothing
        if reads == []:
            results['response1'] = None
            self.results[domain] = results
            return results
        response = reads[0].recvfrom(4096)[0]
        results['response1'] = b64encode(response)
        resp = dns.message.from_wire(response)
        results['response1-ips'] = self.parse_out_ips(resp)

        # if we have made it this far, then wait for the next response
        reads, _, _ = select.select([sock], [], [], self.timeout)
        # if we didn't get anything, then set the results to nothing
        if reads == []:
            results['response2'] = None
            self.results[domain] = results
            return results
        response = reads[0].recvfrom(4096)[0]
        results['response2'] = b64encode(response)
        resp = dns.message.from_wire(response)
        results['response2-ips'] = self.parse_out_ips(resp)
        self.results[domain] = results
        return results

    def parse_out_ips(self, message):
        """Given a message, parse out the ips in the answer"""

        ips = []
        for entry in message.answer:
            for rdata in entry.items:
                ips.append(rdata.to_text())
        return ips



from selenium import webdriver
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.firefox.firefox_binary import FirefoxBinary
from selenium.webdriver.common.proxy import *
from selenium.webdriver import ActionChains
from selenium.common.exceptions import TimeoutException
from selenium.common.exceptions import NoAlertPresentException
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pyvirtualdisplay import Display
import foctor_core.foctor_core as fc
from functools import wraps
import sys
import time
import os
import signal
import errno
import json
import random
import threading
import subprocess


class HeadlessBrowser:
    def __init__(self):
        self.display = Display(visible=False)
        self.binary = None
        self.profile = None
        self.driver = None
        self.parsed = 0

    @fc.timing
    def setup_profile(self, firebug=True, netexport=True):
        """
        Setup the profile for firefox
        :param firebug: whether add firebug extension
        :param netexport: whether add netexport extension
        :return: a firefox profile object
        """
        capture_path = os.getcwd()
        profile = webdriver.FirefoxProfile()
        profile.set_preference("app.update.enabled", False)
        if firebug:
            profile.add_extension(capture_path + '/extensions/firebug-2.0.8.xpi')
            profile.set_preference("extensions.firebug.currentVersion", "2.0.8")
            profile.set_preference("extensions.firebug.allPagesActivation", "on")
            profile.set_preference("extensions.firebug.defaultPanelName", "net")
            profile.set_preference("extensions.firebug.net.enableSites", True)
            profile.set_preference("extensions.firebug.delayLoad", False)
            profile.set_preference("extensions.firebug.onByDefault", True)
            profile.set_preference("extensions.firebug.showFirstRunPage", False)
            profile.set_preference("extensions.firebug.net.defaultPersist", True)  # persist all redirection responses
        if netexport:
            fc.make_folder(capture_path + "/har/")
            profile.add_extension(capture_path + '/extensions/netExport-0.9b7.xpi')
            profile.set_preference("extensions.firebug.DBG_NETEXPORT", True)
            profile.set_preference("extensions.firebug.netexport.alwaysEnableAutoExport", True)
            profile.set_preference("extensions.firebug.netexport.defaultLogDir", capture_path + "/har/")
            profile.set_preference("extensions.firebug.netexport.includeResponseBodies", True)
        return profile

    def open_virtual_display(self):
        self.display.start()

    def close_virtual_display(self):
        self.display.stop()

    def wrap_results(self, **kwargs):
        """
        Wrap returned http response into a well formatted dict
        :param kwargs (dict): this param should contains following keys:
                                fd: file directory to 
                                url: the test url fo the result
                                files_count: the number of files under har/ directory

        :return (dict): the results of all 
        """

        if 'fd' not in kwargs or \
                        'url' not in kwargs or \
                        'files_count' not in kwargs:
            sys.stderr.write("not enough arguments in wrap_results function\n")
            return {}

        external = kwargs['external'] if 'external' in kwargs else None
        fd = kwargs['fd']
        url = kwargs['url']
        length = kwargs['files_count']

        results = {}

        files = []
        wait_time = 15
        host = self.divide_url(url)[0]
        print host
        time.sleep(0.5)

        # wait until the har file is generated
        while len(os.listdir(fd)) <= length + self.parsed:
            time.sleep(1)
            wait_time -= 1
            if wait_time == 0:
                print "\r%s waiting har file result timeout\n" % url
                results['error'] = "wrap har file timeout"
                if external is not None:
                    external[url] = results
                return results
            sys.stdout.write("\rwaiting for har file... %ds" % (wait_time))
            sys.stdout.flush()
        print ''
        time.sleep(1)

        # find all har files under har/ directory
        for fn in os.listdir(fd):
            if fn.endswith(".har") and fn.startswith(host):
                path = fd + '/' + fn
                files.append((fn, os.stat(path).st_mtime))

        # sort all har files and parse the latest one
        files.sort(key=lambda x: x[1])
        with open(fd + '/' + files[-1][0]) as f:
            raw_data = json.load(f)['log']['entries']
            results = [{} for i in range(0, len(raw_data))]
            for i in range(0, len(results)):

                results[i]['request'] = {}
                results[i]['request']['method'] = raw_data[i]['request']['method']
                headers = {}
                for header in raw_data[i]['request']['headers']:
                    headers[header['name']] = header['value']
                results[i]['request']['headers'] = headers

                results[i]['response'] = {}
                results[i]['response']['status'] = raw_data[i]['response']['status']
                results[i]['response']['reason'] = raw_data[i]['response']['statusText']
                headers = {}
                for header in raw_data[i]['response']['headers']:
                    headers[header['name']] = header['value']
                results[i]['response']['headers'] = headers
                results[i]['response']['redirect'] = raw_data[i]['response']['redirectURL']
                results[i]['response']['body'] = raw_data[i]['response']['content']

                sys.stdout.write("\rstart parsing %s %d%%" % (url, int(100 * (i + 1) / len(results))))
                sys.stdout.flush()
            print "\n"

        self.parsed += 1  # increment the number of parsed har files
        sys.stdout.write('\n\n')
        sys.stdout.flush()

        # save test result of this url to the external result object or 
        # return the result
        if external is not None:
            external[url] = results
        else:
            return results

    def divide_url(self, url):
        """
        divide url into host and path two parts
        """
        if 'https://' in url:
            host = url[8:].split('/')[0]
            path = url[8 + len(host):]
        elif 'http://' in url:
            host = url[7:].split('/')[0]
            path = url[7 + len(host):]
        else:
            host = url.split('/')[0]
            path = url[len(host):]
        return host, path

    def get(self, host, files_count, path="/", ssl=False, external=None):
        """
        Send get request to a url and wrap the results
        :param host (str): the host name of the url
        :param path (str): the path of the url (start with "/")
        :return (dict): the result of the test url
        """
        theme = "https" if ssl else "http"
        url = host + path
        http_url = theme + "://" + url

        result = {}
        try:
            capture_path = os.getcwd() + '/'

            har_file_path = capture_path + "har/"

            # fc.load_page(self.driver, http_url)
            fc.switch_tab(self.driver)
            self.load_page(http_url)

            print "driver get: " + http_url

            time.sleep(2)

            if url[-1] == "/":
                f_name = url.split('/')[-2]
            else:
                f_name = url.split('/')[-1]

            fc.save_html(self.driver, f_name, capture_path + "htmls/")
            fc.save_screenshot(self.driver, f_name, capture_path + "screenshots/")

            result = self.wrap_results(url=http_url, files_count=files_count, fd=har_file_path)

            if external is not None:
                external[http_url] = result

        except Exception as e:
            result['error'] = e.message
            print e

        return result

    def foctor_get_requests(self, site_list_file, results):
        """
        use foctor_core library do get requests
        site_list_file (str): the file name of the list of test urls
                        format: 
                            1, www.facebook.com
                            2, www.google.com
                            ...
        results (dict): the object to save the responses from server
        """

        capture_path = os.getcwd() + "/"
        fc.make_folder(capture_path)

        display_mode = 0  # 0 is virtural display(Xvfb mode)

        site_list = fc.read_site_list(site_list_file=site_list_file, start_index=0, end_index=1 << 31)

        driver, display = fc.do_crawl(sites=site_list, driver=self.driver, display=self.display,
                                      capture_path=capture_path, callback=self.wrap_results,
                                      external=results, fd=capture_path + "har/",
                                      files_count=len(os.listdir(capture_path + "har/")))
        fc.teardown_driver(driver, display, display_mode)

        driver.quit()  # quit driver will also clean up the tmp file under /tmp directory

    def run(self, input_file, url=None, verbose=0):
        """
        run the headless browser with given input
        if url given, the proc will only run hlb with given url and ignore input_list.
        :param url:
        :param input_list:
        :param input_file: the name of the file in "index url" format. i.e.
                1, www.facebook.com
                1, www.google.com
                ...
        :return:
        """
        if not url and not input_file:
            print 'no inputs'
            return {"error": "no inputs"}

        results = {}

        self.open_virtual_display()
        if verbose > 0:
            log_file = sys.stdout
        else:
            log_file = None

        # set up firefox driver 
        self.binary = FirefoxBinary(log_file=log_file)
        self.profile = self.setup_profile()
        self.driver = webdriver.Firefox(firefox_profile=self.profile, firefox_binary=self.binary, timeout=60)
        self.driver.set_page_load_timeout(60)
        fc.make_folder(os.getcwd() + "/har/")

        isfile = False
        if url:
            host, path = self.divide_url(url)
            results[url] = self.get(host, path)
        else:
            self.foctor_get_requests(input_file, results)
            isfile = True

        # foctor_core will quit the driver by itself so we only quit the driver when we don't use foctor core
        if not isfile:
            print "quit driver"
            self.quit()
            self.close_virtual_display()

        with open("./hb_results.json", "w") as f:
            json.dump(results, f, indent=4)
        print "Headless browser test finished"
        return results
